import time

from django_websocket_server.messagetemplates import commandmessage
from django_websocket_server.websocket_server import SocketServer
from multi_purpose_arduino_controller.arduino_controller.api import ArduinoControllerAPI



class BoardControllerAPI(ArduinoControllerAPI):
    def __init__(self, **kwargs):
        kwargs.setdefault("name", "BoardControllerAPI")
        super().__init__(**kwargs)


    def port_opened(self,port):
        super().port_opened(port)
        if self.websocket_server is not None:
            self.websocket_server.send_to_all(commandmessage(cmd="port_opened",port=port,sender=self.name))

    def port_closed(self,port):
        super().port_closed(port)
        if self.websocket_server is not None:
            self.websocket_server.send_to_all(commandmessage(cmd="port_closed",port=port,sender=self.name))

    def ws_get_ports(self):
        self._python_communicator.cmd_out(cmd="get_ports",targets=["serialreader"],data_target=self.name)
        time.sleep(0.5)
        if self.websocket_server is not None:
            self.websocket_server.send_to_all(
                commandmessage(cmd="set_ports",connected_ports = list(self.connected_ports),ignored_ports = list(self.ignored_ports),available_ports=list(self.available_ports),identified_ports=list(self.identified_ports),sender=self.name)
            )

    def port_identified(self,port):
        super().port_identified(port)
        self.ws_get_ports()

    def activate_port(self,port):
        self._python_communicator.cmd_out(cmd="reactivate_port",targets=["serialreader"],port=port)

    def deactivate_port(self,port):
        self._python_communicator.cmd_out(cmd="deactivate_port",targets=["serialreader"],port=port)

    def get_board(self,port):
        self._python_communicator.cmd_out(cmd="get_board",targets=[port],data_target=self.name)

    def set_board(self,board,port):
        self.websocket_server.send_to_all(
            commandmessage(cmd="set_board",board=board,sender=self.name,port=port)
        )

    def set_board_attribute(self,port,attribute,value):
        self._python_communicator.cmd_out(cmd="set_board_attribute",targets=[port],attribute=attribute,value=value)

    def set_websocket_server(self, websocket_server:SocketServer):
        super().set_websocket_server(websocket_server)
        self.websocket_server.register_cmd("get_ports",self.ws_get_ports)
        self.websocket_server.register_cmd("activate_port",self.activate_port)
        self.websocket_server.register_cmd("deactivate_port",self.deactivate_port)
        self.websocket_server.register_cmd("get_board",self.get_board)
        self.websocket_server.register_cmd("set_board_attribute",self.set_board_attribute)




from django import forms
from django.forms import ModelForm
from django.shortcuts import render


# Create your views here.
def index(request):
    return render(request, "index.html")


def read_only_form(form: ModelForm):
    for key in form.fields.keys():
        if isinstance(form.fields[key].widget, forms.Select):
            form.fields[key].widget = forms.TextInput()
        form.fields[key].widget.attrs["readonly"] = True

    return form



import logging
import threading
import time

import serial

from .parseboards import board_by_firmware
from multi_purpose_arduino_controller.arduino_controller.serialreader.serialreader import SerialReader
from .basicboard.board import ArduinoBasicBoard
from .portrequest import validate_buffer
from python_communicator import PythonCommunicator

PORTREADTIME = 0.01
BAUDRATES = (
    9600,
    115200,
    #   19200,
    #   38400,
    #   57600,
    #    230400,
    #    460800,
    #    500000,
    #    576000,
    #    921600,
    #    1000000,
    #    1152000,
    #    1500000,
    #    2000000,
    #    2500000,
    #    3000000,
    #    3500000,
    #    4000000,
)


class SerialPort(serial.Serial):
    def __init__(
        self,
        serialreader,
        port,
        config=None,
        logger=None,
        communicator: PythonCommunicator = None,
        baudrate=9600,
        **kwargs,
    ):

        self.serialreader: SerialReader = serialreader
        self.data_targets = set([])

        if logger is None:
            logger = logging.getLogger("serialreader " + port)
        self.logger = logger

        self.board: ArduinoBasicBoard = None
        self.workthread = None
        self.updatethread = None
        self.readbuffer = []
        self.time = time.time()

        if config is None:
            config = serialreader.config
        self.config = config

        try:
            super().__init__(port, baudrate=baudrate, timeout=0, **kwargs)
        except Exception as e:
            serialreader.deadports.add(port)
            self.logger.exception(e)
            return
        self.logger.info("port found " + self.port)
        serialreader.connected_ports.add(self)

        if communicator is None:
            communicator = PythonCommunicator()
        self.communicator:PythonCommunicator = communicator

        self.communicator.add_node(port, self)

        self.to_write = []
        self.start_read()

        newb = board_by_firmware(config.get("portdata", self.port, "firmware", default=0))
        if newb is not None:
            self.set_board(newb["classcaller"])
        else:
            self.set_board(ArduinoBasicBoard)

    def add_data_target(self, data_target=None):
        if data_target is not None:
            self.data_targets.add(data_target)

    def remove_data_target(self, data_target):
        if data_target in self.data_targets:
            self.data_targets.remove(data_target)

    def boardfunction(self, board_cmd, **kwargs):
        try:
            getattr(self.board, board_cmd)(**kwargs)
        except:
            self.logger.exception(Exception)

    def set_board_attribute(self, attribute,value):
        try:
            setattr(self.board,attribute,value)
        except:
            self.logger.exception(Exception)

    def add_datapoint(self,board, key, y, x=None):
        t = time.time() - self.time
        if x is None:
            x = t
        self.serialreader.datalogger.add_datapoint(key, x=x, y=y)
        self.communicator.cmd_out(cmd="datapoint", key=key, x=x, y=y, targets=list(self.data_targets),port=board.port)

    def set_board(self, BoardClass):
        self.board = BoardClass()
        self.board.set_serial_port(self)
        time.sleep(2)
        self.board.identify()

        if not self.board.identified:
            self.stop_read()
            self.serialreader.ignored_ports.add(self.port)
            raise ValueError("unable to identify " + self.port)

        if self.board.FIRMWARE != self.board.firmware:
            newb = board_by_firmware(self.board.firmware)
            if newb is not None:
                return self.set_board(newb["classcaller"])
            else:
                self.stop_read()
                self.serialreader.ignored_ports.add(self.port)
                raise ValueError("firmware not found " + str(self.board.firmware))

        #self.board.specific_identification()
        if not self.board.identified:
            self.stop_read()
            self.serialreader.ignored_ports.add(self.port)
            raise ValueError(
                "unable to specificidentify "
                + self.port
                + "with firmware:"
                + str(self.board.firmware)
            )

        self.logger.info(str(self.port) + " identified ")

        self.serialreader.identified_ports.add(self)

        self.config.put("portdata", self.port, "baud", value=self.baudrate)
        self.config.put("portdata", self.port, "firmware", value=self.board.firmware)
        self.board.restore(self.config.get("boarddata", self.board.id, default={}))
        self.board.get_portcommand_by_name("identify").sendfunction(True)
        self.communicator.cmd_out(cmd="port_identified",port=self.port)
        return True

    def board_updater(self):
        while self.is_open:
            if self.board is not None:
                self.send_board_data()
                time.sleep(self.board.updatetime)

    def send_board_data(self, **kwargs):
        if self.board is not None:
            if self.board.identified:
                data = self.board.save()
                self.config.put("boarddata", self.board.id, value=data)
                msg_d = dict(
                    targets=list(self.data_targets),
                    cmd="boardupdate",
                    boarddata=data,
                    **kwargs,
                )
                msg = self.communicator.cmd_out(**msg_d)
                self.logger.info(msg_d)

    def work_port(self):
        while self.is_open:
            try:
                while len(self.to_write) > 0:
                    t = self.to_write.pop()
                    self.logger.debug("write to " + self.port + ": " + str(t))
                    super().write(t)
                c = self.read()

                while len(c) > 0:
                    self.readbuffer.append(c)
                    validate_buffer(self)
                    c = self.read()
            except Exception as e:
                self.logger.exception(e)
                break
            time.sleep(PORTREADTIME)
        self.logger.error("work_port stopped")
        self.stop_read()

    def start_read(self):
        self.logger.info("port opened " + self.port)
        self.communicator.cmd_out(
            sender=self.port, cmd="port_opened",port=self.port,baud=self.baudrate)

        if not self.is_open:
            self.open()
        self.workthread = threading.Thread(target=self.work_port)
        self.updatethread = threading.Thread(target=self.board_updater)
        self.workthread.start()
        self.updatethread.start()

    def write(self, data):
        self.to_write.append(data)

    def stop_read(self):
        self.close()
        try:
            self.workthread.join()
        except:
            pass
        self.workthread = None
        try:
            self.updatethread.join()
        except:
            pass
        self.close()
        self.updatethread = None
        self.logger.info("port closed " + self.port)
        self.communicator.cmd_out(
            sender=self.port,
            cmd="port_closed",
            port = self.port
        )
        if self in self.serialreader.connected_ports:
            self.serialreader.connected_ports.remove(self)
        if self in self.serialreader.identified_ports:
            self.serialreader.identified_ports.remove(self)
        del self

    #def update(self, **kwargs):
    #    if self.board is not None:
    #        if self.board.id is not None:
    #            self.board.restore(**kwargs)
    #           self.send_board_data(force_update=True)

    def get_board(self,data_target=None):
        if data_target is None: return
        board = self.board.get_board()
        self.communicator.cmd_out("set_board",targets=[data_target],board=board,port=self.port)




DEFAULTCODES = {
    "setuint16_t": "uint16_t temp;memcpy(&temp,data,2);{NAME}=temp;write_data({NAME},{BYTEID});",
    "setuint32_t": "uint32_t temp;memcpy(&temp,data,4);{NAME}=temp;write_data({NAME},{BYTEID});",
    "setuint32_t_silent": "uint32_t temp;memcpy(&temp,data,4);{NAME}=temp;",
    "setpin_output": "if(data[0]>0){{NAME}=data[0];}write_data({NAME},{BYTEID});pinMode({NAME}, OUTPUT);",
    "setpin_input": "if(data[0]>0){{NAME}=data[0];}write_data({NAME},{BYTEID});pinMode({NAME}, INPUT);",
    "setbool": "{NAME}=(bool)data[0];write_data({NAME},{BYTEID});",
}



import sys

import glob
import importlib.util
import inspect
import os
from os.path import basename

from .basicboard.board import ArduinoBasicBoard

BOARDS = {}


def board_by_firmware(firmware):
    return BOARDS.get(firmware, None)


def parse_path_for_boards(path):
    global BOARDS
    dir_path = path

    boardsfolders = [
        p
        for p in glob.glob(dir_path + "/*")
        if os.path.isdir(p) and not p.endswith("__")
    ]
    loaded_firmwares = set()
    for boardfolder in boardsfolders:
        boardpy = os.path.join(boardfolder, "board.py")
        if os.path.exists(boardpy):
            spec = importlib.util.spec_from_file_location(
                "board." + basename(boardfolder), os.path.join(boardfolder, "board.py")
            )
            foo = importlib.util.module_from_spec(spec)
            spec.loader.exec_module(foo)
            for name, obj in inspect.getmembers(foo):
                if inspect.isclass(obj):
                    try:
                        fw = getattr(obj, "FIRMWARE")
                        if fw not in loaded_firmwares:
                            BOARDS[fw] = {
                                "firmware": fw,
                                "classcaller": obj,
                                "name": name,
                            }
                            loaded_firmwares.add(fw)
                    except:
                        pass
    print(path)
    print(BOARDS)


parse_path_for_boards(
    os.path.dirname(
        os.path.dirname(
            os.path.abspath(sys.modules[ArduinoBasicBoard.__module__].__file__)
        )
    )
)



import os
import random
import logging

from django.conf import settings
from json_dict import JsonDict

preamble = ""
#if manage.py is called directly
if len(__name__.split(".")) == 2:
    from manage import logger,CONFIG
else:
    from ..manage import logger,CONFIG
    preamble = __name__.replace(".controll_server.settings",".")

BASE_DIR = os.path.dirname(os.path.dirname(os.path.abspath(__file__)))

if CONFIG is None:
    CONFIG = JsonDict(os.path.join(BASE_DIR, "serverconfig.json"))


SECRET_KEY = CONFIG.get(
    "django_settings",
    "security",
    "key",
    default="".join(
        random.SystemRandom().choice(
            "abcdefghijklmnopqrstuvwxyz0123456789!@#$%^&*(-_=+)"
        )
        for i in range(50)
    ),
)

DEBUG = CONFIG.get("django_settings", "debug", default=False)

ALLOWED_HOSTS = CONFIG.get("django_settings", "security", "allowed_hosts", default=[])


LANGUAGE_CODE = CONFIG.get("public", "language_time", "language_code", default="en-us")

TIME_ZONE = CONFIG.get("public", "language_time", "time_zone", default="UTC")

USE_I18N = CONFIG.get("django_settings", "language_time", "use_I18N", default=True)

USE_L10N = CONFIG.get("django_settings", "language_time", "use_L10N", default=True)

USE_TZ = CONFIG.get("django_settings", "language_time", "use_TZ", default=True)

STATIC_URL = CONFIG.get("django_settings", "static_files", "url", default="/static/")


INSTALLED_APPS = []

# APP_DIR=os.path.join(BASE_DIR, 'django_apps')
# if CONFIG.get("django_settings","apps","load_local",default=True):
#    for dir in [f for f in os.listdir(APP_DIR) if os.path.isdir(os.path.join(APP_DIR, f))]:
#        INSTALLED_APPS.append(os.path.basename(APP_DIR)+"."+dir)


if CONFIG.get("django_settings", "apps", "load_defaults", default=True):
    INSTALLED_APPS += [
        "django.contrib.admin",
        "django.contrib.auth",
        "django.contrib.contenttypes",
        "django.contrib.sessions",
        "django.contrib.messages",
        "django.contrib.staticfiles",
        'bootstrap4',
    ]

INSTALLED_APPS += CONFIG.get("django_settings", "apps", "additional", default=[])

STATICFILES_DIRS = [os.path.join(BASE_DIR, "static"), "/var/www/static/"] + CONFIG.get(
    "django_settings", "static_files", "dirs", default=[]
)


MIDDLEWARE = [
    "django.middleware.security.SecurityMiddleware",
    "django.contrib.sessions.middleware.SessionMiddleware",
    "django.middleware.common.CommonMiddleware",
    "django.middleware.csrf.CsrfViewMiddleware",
    "django.contrib.auth.middleware.AuthenticationMiddleware",
    "django.contrib.messages.middleware.MessageMiddleware",
    "django.middleware.clickjacking.XFrameOptionsMiddleware",
    "global_login_required.GlobalLoginRequiredMiddleware",
]

PUBLIC_PATHS = [r"^/accounts/.*"]  # allow public access to all django-allauth views
ROOT_URLCONF = preamble + "controll_server.urls"

TEMPLATES = [
    {
        "BACKEND": "django.template.backends.django.DjangoTemplates",
        "DIRS": [os.path.join(BASE_DIR, "templates")],
        "APP_DIRS": True,
        "OPTIONS": {
            "context_processors": [
                "django.template.context_processors.debug",
                "django.template.context_processors.request",
                "django.contrib.auth.context_processors.auth",
                "django.contrib.messages.context_processors.messages",
                preamble + "templatetags.installed_apps.get_apps_context",
            ],
            "libraries": {"public_dict": preamble + "templatetags.public_dict"},
        },
    }
]

WSGI_APPLICATION = preamble + "controll_server.wsgi.application"


# Database
# https://docs.djangoproject.com/en/2.2/ref/settings/#databases
DATABASES = {
    "default": {
        "ENGINE": "django.db.backends.sqlite3",
        "NAME":  CONFIG.get(
            "django_settings",
            "database",
            "name",
            default=os.path.abspath(os.path.join(os.path.dirname(CONFIG.file), "db.sqlite3")),
        )
    }
}
logger.debug("use db:"+str(DATABASES['default']))


# Password validation
# https://docs.djangoproject.com/en/2.2/ref/settings/#auth-password-validators

AUTH_PASSWORD_VALIDATORS = [
    {
        "NAME": "django.contrib.auth.password_validation.UserAttributeSimilarityValidator"
    },
    {"NAME": "django.contrib.auth.password_validation.MinimumLengthValidator"},
    {"NAME": "django.contrib.auth.password_validation.CommonPasswordValidator"},
    {"NAME": "django.contrib.auth.password_validation.NumericPasswordValidator"},
]



i = -1
if __name__.split(".")[i] == "django_app":
    i = -2

default_app_config = (
        __name__
        + ".apps."
        + "".join(x.capitalize() or "_" for x in __name__.split(".")[i].split("_"))
        + "Config"
)
